# Copyright (C) 2014-2016 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Subject header prefix munging."""

import re

from contextlib import suppress
from email.header import Header, decode_header, make_header
from mailman import public
from mailman.core.i18n import _
from mailman.interfaces.handler import IHandler
from zope.interface import implementer


RE_PATTERN = '\s*((RE|AW|SV|VS)(\[\d+\])?\s*:\s*)+'
ASCII_CHARSETS = (None, 'ascii', 'us-ascii')
EMPTYSTRING = ''


def get_subject(mlist, msgdata, subject, prefix, prefix_pattern, ws):
    list_charset = mlist.preferred_language.charset
    chunks = decode_header(subject.encode())
    # Only search the first chunk for Re and existing prefix.
    chunk_text, chunk_charset = chunks[0]
    if chunk_charset is None:
        chunk_charset = 'us-ascii'
    if isinstance(chunk_text, bytes):
        chunk_text = chunk_text.decode(chunk_charset)
    chunk_text = re.sub(prefix_pattern, '', chunk_text)
    if not chunk_text.strip():
        with _.using(mlist.preferred_language.code):
            subject_text = _('(no subject)')
        chunks = [(prefix.strip(), list_charset),
                  (subject_text, list_charset),
                  ]
        msgdata['stripped_subject'] = str(make_header(
            chunks[1:], continuation_ws=ws))
        return make_header(chunks, continuation_ws=ws)
    first_text = re.sub(prefix_pattern, '', chunk_text).lstrip()
    rematch = re.match(RE_PATTERN, first_text, re.I)
    if rematch:
        first_text = 'Re: ' + first_text[rematch.end():]
    chunks[0] = (first_text, chunk_charset)
    # The subject text stripped of the prefix, for use in the NNTP gateway.
    msgdata['stripped_subject'] = str(make_header(chunks, continuation_ws=ws))
    chunks.insert(0, (prefix.strip(), list_charset))
    return make_header(chunks, continuation_ws=ws)


@public
@implementer(IHandler)
class SubjectPrefix:
    """Add a list-specific prefix to the Subject header value."""

    name = 'subject-prefix'
    description = _('Add a list-specific prefix to the Subject header value.')

    def process(self, mlist, msg, msgdata):
        """See `IHandler`."""
        if msgdata.get('isdigest') or msgdata.get('_fasttrack'):
            return
        prefix = mlist.subject_prefix
        if not prefix.strip():
            return
        subject = msg.get('subject', '')
        # Turn the value into a Header instance and try to figure out what
        # continuation whitespace is being used.
        # Save the original Subject.
        msgdata['original_subject'] = subject
        if isinstance(subject, Header):
            subject_text = str(subject)
        else:
            subject = make_header(decode_header(subject))
            subject_text = str(subject)
        lines = subject_text.splitlines()
        ws = '\t'
        if len(lines) > 1 and lines[1] and lines[1][0] in ' \t':
            ws = lines[1][0]
        # If the subject_prefix contains '%d', it is replaced with the mailing
        # list's sequence number.  The sequential number format allows '%d' or
        # '%05d' like pattern.
        prefix_pattern = re.escape(prefix)
        # Unescape '%'.
        prefix_pattern = '%'.join(prefix_pattern.split(r'\%'))
        p = re.compile('%\d*d')
        if p.search(prefix, 1):
            # The prefix has number, so we should search prefix w/number in
            # subject.  Also, force new style.
            prefix_pattern = p.sub(r'\s*\d+\s*', prefix_pattern)
        # Substitute %d in prefix with post_id
        with suppress(TypeError):
            prefix = prefix % mlist.post_id
        del msg['subject']
        list_charset = mlist.preferred_language.charset
        chunks = decode_header(subject.encode())
        # Only search the first chunk for Re and existing prefix.
        chunk_text, chunk_charset = chunks[0]
        if chunk_charset is None:
            chunk_charset = 'us-ascii'
        if isinstance(chunk_text, bytes):
            chunk_text = chunk_text.decode(chunk_charset)
        chunk_text = re.sub(prefix_pattern, '', chunk_text)
        if not chunk_text.strip():
            with _.using(mlist.preferred_language.code):
                subject_text = _('(no subject)')
            chunks = [(prefix.strip(), list_charset),
                      (subject_text, list_charset),
                      ]
            msgdata['stripped_subject'] = str(make_header(
                chunks[1:], continuation_ws=ws))
            msg['Subject'] = make_header(chunks, continuation_ws=ws)
            return
        first_text = re.sub(prefix_pattern, '', chunk_text).lstrip()
        rematch = re.match(RE_PATTERN, first_text, re.I)
        if rematch:
            first_text = 'Re: ' + first_text[rematch.end():]
        chunks[0] = (first_text, chunk_charset)
        # The subject text stripped of the prefix, for use in the NNTP gateway.
        msgdata['stripped_subject'] = str(make_header(
            chunks, continuation_ws=ws))
        chunks.insert(0, (prefix.strip(), list_charset))
        msg['Subject'] = make_header(chunks, continuation_ws=ws)
